"""
Fastlane bot configuration object -- selector constants
"""

NETWORK_ETHEREUM = "ethereum"
NETWORK_MAINNET = NETWORK_ETHEREUM
NETWORK_TENDERLY = "tenderly"

DATABASE_SQLITE = "sqlite"
DATABASE_POSTGRES = "postgres"
DATABASE_MEMORY = "memory"
DATABASE_SDK = "sdk"
DATABASE_UNITTEST = "unittest"

LOGGER_DEFAULT = "default"

LOGLEVEL_DEBUG = "debug"
LOGLEVEL_INFO = "info"
LOGLEVEL_WARNING = "warning"
LOGLEVEL_ERROR = "error"

PROVIDER_DEFAULT = "default"  # the default provider for network
PROVIDER_INFURA = "infura"
PROVIDER_ALCHEMY = "alchemy"
PROVIDER_TENDERLY = "tenderly"
PROVIDER_UNITTEST = "unittest"
