
__VERSION__ = "0.0"
__DATE__="01/May/2023"

class TxReceiptHandlerBase:
    __VERSION__=__VERSION__
    __DATE__=__DATE__

class TxReceiptHandler(TxReceiptHandlerBase):
    __VERSION__=__VERSION__
    __DATE__=__DATE__